
;RHL Window - DednDave, 4-2012

        .XCREF
        .NOLIST
        INCLUDE    \masm32\include\masm32rt.inc
        INCLUDE    RHLWin.inc
        .LIST

;###############################################################################################

MainWinWidth  EQU 400
MainWinHeight EQU 152

;###############################################################################################

        .DATA
        ALIGN   4

;-----------------------------------------------------------------------------------------------
;
;  When we have a program that creates a number of controls and/or windows at various times
;throughout execution, it may be advantageous to use tables to define them and a single looping
;routine to perform the creation. This makes the creation code easier to follow or modify by
;centralizing the creation variables and can also reduce overall code/data size.
;
;  In this implementation, each table begins with a "header" containing two DWORD's:
;
;    lpHandleList:LPVOID
;      uItemCount:UINT
;
;  The lpHandleList value may be used to point to a list or structure that will receive handles
;as the windows or controls are created. If used, the list must contain at least as many DWORD
;entries as there are windows/controls in the table. If this value is NULL, the handles are not
;stored. This may be desirable when creating a window, where the hWin value is stored in the
;WM_CREATE handler, as the handle is available at an earlier point in execution. The uItemCount
;value specifies the number of windows or controls in the table.
;
;  The header is followed by a list of structures for each window or control to be created.
;
;CreateWindowEx parameters:
;
;       dwExStyle:DWORD
;     lpClassName:LPCTSTR
;    lpWindowName:LPCTSTR
;         dwStyle:DWORD
;               x:UINT
;               y:UINT
;          nWidth:UINT
;         nHeight:UINT
;      hWndParent:HWND
;           hMenu:HMENU
;       hInstance:HINSTANCE
;         lpParam:LPVOID
;
;  We want to call CreateWindowEx in a loop to create sets of controls or windows. Two of the
;required parameters are constant and will be in registers during the loop. For any given set
;of windows, the constant parameters will be:
;
;      hWndParent:HWND      (handle to the parent window, NULL for main window)
;       hInstance:HINSTANCE (always hInstance)
;
;  The other parameters are different (variable) for each window and are placed in the table.
;As we pass through the loop, parameters are taken from the table and PUSH'ed onto the stack.
;Notice that parameters are PUSH'ed in reverse order, so we create our tables that way.
;
;window/control creation table parameters:
;
;         lpParam:LPVOID
;           hMenu:HMENU
;         nHeight:UINT
;          nWidth:UINT
;               y:UINT
;               x:UINT
;         dwStyle:DWORD
;    lpWindowName:LPCTSTR
;     lpClassName:LPCTSTR
;       dwExStyle:DWORD
;
; 1) For controls, hMenu is the control ID and lParam should be 0
; 2) For non-MDI windows, lParam may be used to point to a child creation table
; 3) For MDI client windows, lParam should point to a CLIENTCREATESTRUCT structure
; 4) For MDI child windows, lParam should point to a MDICREATESTRUCT structure
; 5) For windows with a title bar, lpWindowName points to a caption string.
; 6) For windows with no title bar, lpWindowName points to a string for text inside the window.
; 7) For static controls with the SS_ICON flag set, lpWindowName may be the icon identifier.
;
;  Whether we are creating windows or controls, each call to CreateWindowEx requires a
;lpClassName value. This parameter determines what type of window or control is to be created.
;While the value may represent a "local class atom", that method is rarely used in assembly
;language. We generally use a pointer to a string that describes the class name.
;
;  For system-defined controls, these classes are already registered and have system-internal
;WndProc routines:
;
;   Control Type              String
;---------------------------------------------------
;   Button Control            'Button'
;   Edit Box                  'Edit'
;   Static Control            'Static'
;   List Box                  'Listbox'
;   Scroll Bar                'Scrollbar'
;   Combo Box                 'ComboBox'
;   Header Control            'SysHeader32'
;   List View                 'SysListView32'
;   Tree View                 'SysTreeView32'
;   Hot Key                   'msctls_hotkey32'
;   Up-Down Control           'msctls_updown32'
;   Animation Control         'SysAnimate32'
;   Extended Combo Box        'ComboBoxEx32'
;   Tab Control               'SysTabControl32'
;   Month Calendar Control    'SysMonthCal32'
;   Progress Bar              'msctls_progress32'
;   ReBar Control             'ReBarWindow32'
;   Tool Tip                  'tooltips_class32'
;   Track Bar                 'msctls_trackbar32'
;   Status Bar                'msctls_statusbar32'
;   Tool Bar                  'ToolbarWindow32'
;   Date/Time Picker          'SysDateTimePick32'
;   IP Address Control        'SysIPAddress32'
;   Page Scroller Control     'SysPager'
;   Drag List Message String  'commctrl_DragListMsg'
;
;There are a few others, but the list represents most of the ones used by programmers today.
;
;  For other window types, like "our own" windows, we must provide a WndProc routine and
;register the window class. These classes must be registered before the string that describes
;them may be used in a call to CreateWindowEx.
;
;-----------------------------------------------------------------------------------------------

MainCtrlsTable  dd MainHandles,3
                dd NULL,IDC_BTN01,26,86,10,10
                dd WS_CHILD or WS_VISIBLE or WS_GROUP or WS_TABSTOP
                dd szButton01,szBtnClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN02,26,86,44,10
                dd WS_CHILD or WS_VISIBLE
                dd szButton02,szBtnClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN03,26,86,78,10
                dd WS_CHILD or WS_VISIBLE
                dd szButton03,szBtnClass,WS_EX_NOPARENTNOTIFY

Type1CtrlsTable dd Type1Handles,4
                dd NULL,IDC_STC1,112,288,0,0
                dd WS_CHILD or WS_VISIBLE or SS_BITMAP
                dd NULL,szStcClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN11,26,86,10,10
                dd WS_CHILD or WS_VISIBLE or WS_GROUP or WS_TABSTOP
                dd szButton11,szBtnClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN12,26,86,44,10
                dd WS_CHILD or WS_VISIBLE
                dd szButton12,szBtnClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN13,26,86,78,10
                dd WS_CHILD or WS_VISIBLE
                dd szButton13,szBtnClass,WS_EX_NOPARENTNOTIFY

Type2CtrlsTable dd Type2Handles,4
                dd NULL,IDC_STC2,112,288,0,0
                dd WS_CHILD or WS_VISIBLE or SS_BITMAP
                dd NULL,szStcClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN21,26,86,10,10
                dd WS_CHILD or WS_VISIBLE or WS_GROUP or WS_TABSTOP
                dd szButton21,szBtnClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN22,26,86,44,10
                dd WS_CHILD or WS_VISIBLE
                dd szButton22,szBtnClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN23,26,86,78,10
                dd WS_CHILD or WS_VISIBLE
                dd szButton23,szBtnClass,WS_EX_NOPARENTNOTIFY

Type3CtrlsTable dd Type3Handles,4
                dd NULL,IDC_STC3,112,288,0,0
                dd WS_CHILD or WS_VISIBLE or SS_BITMAP
                dd NULL,szStcClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN31,26,86,10,10
                dd WS_CHILD or WS_VISIBLE or WS_GROUP or WS_TABSTOP
                dd szButton31,szBtnClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN32,26,86,44,10
                dd WS_CHILD or WS_VISIBLE
                dd szButton32,szBtnClass,WS_EX_NOPARENTNOTIFY
                dd NULL,IDC_BTN33,26,86,78,10
                dd WS_CHILD or WS_VISIBLE
                dd szButton33,szBtnClass,WS_EX_NOPARENTNOTIFY

Type1WinTable   dd NULL,1
                dd Type1CtrlsTable,IDC_TYPE1,114,290,2,100
                dd WS_CHILD or WS_VISIBLE or WS_GROUP or WS_TABSTOP or WS_BORDER or WS_CLIPCHILDREN
                dd NULL,szType1Class,NULL

Type2WinTable   dd NULL,1
                dd Type2CtrlsTable,IDC_TYPE2,114,290,2,100
                dd WS_CHILD or WS_VISIBLE or WS_GROUP or WS_TABSTOP or WS_BORDER or WS_CLIPCHILDREN
                dd NULL,szType2Class,NULL

Type3WinTable   dd NULL,1
                dd Type3CtrlsTable,IDC_TYPE3,114,290,2,100
                dd WS_CHILD or WS_VISIBLE or WS_GROUP or WS_TABSTOP or WS_BORDER or WS_CLIPCHILDREN
                dd NULL,szType3Class,NULL

szAppName       db 'RHL Window',0
szMainClass     db 'RHLMainClass',0
szType1Class    db 'RHLType1Class',0
szType2Class    db 'RHLType2Class',0
szType3Class    db 'RHLType3Class',0
szStcClass      db 'Static',0
szBtnClass      db 'Button',0

szButton01      db 'Type 1',0
szButton02      db 'Type 2',0
szButton03      db 'Type 3',0

szButton11      db 'Button 1',0
szButton12      db 'Button 2',0
szButton13      db 'Button 3',0

szButton21      db 'Button 4',0
szButton22      db 'Button 5',0
szButton23      db 'Button 6',0

szButton31      db 'Button 7',0
szButton32      db 'Button 8',0
szButton33      db 'Button 9',0

;------------------------------

        .DATA?
        ALIGN   4

hInstance       dd ?
hFont           dd ?
hWin            dd ?
hWin1           dd ?
hWin2           dd ?
hWin3           dd ?
hCurrentWindow  dd ?
nCurrentBtnID   dd ?

MainHandles     LABEL DWORD
hButton01       dd ?
hButton02       dd ?
hButton03       dd ?

Type1Handles    LABEL DWORD
hStatic1        dd ?
hButton11       dd ?
hButton12       dd ?
hButton13       dd ?
hImage1         dd ?

Type2Handles    LABEL DWORD
hStatic2        dd ?
hButton21       dd ?
hButton22       dd ?
hButton23       dd ?
hImage2         dd ?

Type3Handles    LABEL DWORD
hStatic3        dd ?
hButton31       dd ?
hButton32       dd ?
hButton33       dd ?
hImage3         dd ?

;###############################################################################################

        .CODE
        ALIGN   16

WndProc PROC    hWnd:HWND,uMsg:UINT,wParam:WPARAM,lParam:LPARAM

;------------------------------

        mov     eax,uMsg
        cmp     eax,WM_GETMINMAXINFO
        jz      WndPr2

        cmp     eax,WM_COMMAND
        jz      WndPr4

        cmp     eax,WM_CREATE
        jz      WndPrX

        cmp     eax,WM_CLOSE
        jz      WndPrY

        cmp     eax,WM_DESTROY
        jz      WndPrZ

WndPr0: INVOKE  DefWindowProc,hWnd,uMsg,wParam,lParam

WndPr1: ret

;------------------------------

;WM_GETMINMAXINFO

WndPr2: mov     edx,lParam
        mov     eax,MainWinWidth
        mov     ecx,MainWinHeight
        mov     [edx].MINMAXINFO.ptMinTrackSize.x,eax
        mov     [edx].MINMAXINFO.ptMinTrackSize.y,ecx
        mov     [edx].MINMAXINFO.ptMaxTrackSize.x,eax
        mov     [edx].MINMAXINFO.ptMaxTrackSize.y,ecx

;------------------------------

;return FALSE

WndPr3: xor     eax,eax
        jmp     WndPr1

;------------------------------

;WM_COMMAND

WndPr4: movzx   eax,word ptr wParam+2
        cmp     ax,BN_CLICKED
        jnz     WndPr3

;re-enable the previous button

        INVOKE  GetDlgItem,hWnd,nCurrentBtnID
        INVOKE  EnableWindow,eax,TRUE

;close the current window

        INVOKE  PostMessage,hCurrentWindow,WM_SYSCOMMAND,SC_CLOSE,NULL

;disable the new button

        movzx   eax,word ptr wParam
        push    eax
        mov     nCurrentBtnID,eax
        INVOKE  GetDlgItem,hWnd,eax
        INVOKE  EnableWindow,eax,FALSE
        pop     eax

;create the new window

        mov     edx,offset Type1WinTable
        cmp     ax,IDC_BTN01
        jz      WndPr5

        mov     edx,offset Type2WinTable
        cmp     ax,IDC_BTN02
        jz      WndPr5

        cmp     ax,IDC_BTN03
        jnz     WndPr3

        mov     edx,offset Type3WinTable

;TableCreate returns the handle of the last window/control created

WndPr5: INVOKE  TableCreate,hWnd,edx
        mov     hCurrentWindow,eax
        jmp     WndPr3

;------------------------------

;WM_CREATE

WndPrX: mov     edx,hWnd
        mov     hWin,edx

;create the window controls

        INVOKE  TableCreate,edx,offset MainCtrlsTable

;create a Type 1 window by simulating a button click

        INVOKE  PostMessage,hWnd,WM_COMMAND,IDC_BTN01+65536*BN_CLICKED,hButton01
        jmp     WndPr3

;------------------------------

;WM_CLOSE

WndPrY: INVOKE  DestroyWindow,hWnd
        jmp     WndPr3

;------------------------------

;WM_DESTROY

WndPrZ: INVOKE  PostQuitMessage,NULL
        jmp     WndPr3

;------------------------------

WndProc ENDP

;***********************************************************************************************

Type1Proc PROC  hWnd:HWND,uMsg:UINT,wParam:WPARAM,lParam:LPARAM

;------------------------------

        mov     eax,uMsg
        cmp     eax,WM_CREATE
        jz      Ty1PrX

        cmp     eax,WM_CLOSE
        jz      Ty1PrY

Ty1Pr0: INVOKE  DefWindowProc,hWnd,uMsg,wParam,lParam

Ty1Pr1: ret

;------------------------------

Ty1Pr2:

;------------------------------

;return FALSE

Ty1Pr3: xor     eax,eax
        jmp     Ty1Pr1

;------------------------------

;WM_CREATE

Ty1PrX: mov     edx,hWnd
        mov     hWin1,edx

;create the window controls

        INVOKE  TableCreate,edx,offset Type1CtrlsTable

;load the background image and set it to the static control

        INVOKE  LoadBitmap,hInstance,IDB_IMG1
        mov     hImage1,eax
        INVOKE  SendMessage,hStatic1,STM_SETIMAGE,IMAGE_BITMAP,eax

;initialize any other controls here

        jmp     Ty1Pr3

;------------------------------

;WM_CLOSE

Ty1PrY: INVOKE  DeleteObject,hImage1
        INVOKE  DestroyWindow,hWnd
        jmp     Ty1Pr3

;------------------------------

Type1Proc ENDP

;***********************************************************************************************

Type2Proc PROC  hWnd:HWND,uMsg:UINT,wParam:WPARAM,lParam:LPARAM

;------------------------------

        mov     eax,uMsg
        cmp     eax,WM_CREATE
        jz      Ty2PrX

        cmp     eax,WM_CLOSE
        jz      Ty2PrY

Ty2Pr0: INVOKE  DefWindowProc,hWnd,uMsg,wParam,lParam

Ty2Pr1: ret

;------------------------------

Ty2Pr2:

;------------------------------

;return FALSE

Ty2Pr3: xor     eax,eax
        jmp     Ty2Pr1

;------------------------------

;WM_CREATE

Ty2PrX: mov     edx,hWnd
        mov     hWin2,edx

;create the window controls

        INVOKE  TableCreate,edx,offset Type2CtrlsTable

;load the background image and set it to the static control

        INVOKE  LoadBitmap,hInstance,IDB_IMG2
        mov     hImage2,eax
        INVOKE  SendMessage,hStatic2,STM_SETIMAGE,IMAGE_BITMAP,eax

;initialize any other controls here

        jmp     Ty2Pr3

;------------------------------

;WM_CLOSE

Ty2PrY: INVOKE  DeleteObject,hImage2
        INVOKE  DestroyWindow,hWnd
        jmp     Ty2Pr3

;------------------------------

Type2Proc ENDP

;***********************************************************************************************

Type3Proc PROC  hWnd:HWND,uMsg:UINT,wParam:WPARAM,lParam:LPARAM

;------------------------------

        mov     eax,uMsg
        cmp     eax,WM_CREATE
        jz      Ty3PrX

        cmp     eax,WM_CLOSE
        jz      Ty3PrY

Ty3Pr0: INVOKE  DefWindowProc,hWnd,uMsg,wParam,lParam

Ty3Pr1: ret

;------------------------------

Ty3Pr2:

;------------------------------

;return FALSE

Ty3Pr3: xor     eax,eax
        jmp     Ty3Pr1

;------------------------------

;WM_CREATE

Ty3PrX: mov     edx,hWnd
        mov     hWin3,edx

;create the window controls

        INVOKE  TableCreate,edx,offset Type3CtrlsTable

;load the background image and set it to the static control

        INVOKE  LoadBitmap,hInstance,IDB_IMG3
        mov     hImage3,eax
        INVOKE  SendMessage,hStatic3,STM_SETIMAGE,IMAGE_BITMAP,eax

;initialize any other controls here

        jmp     Ty3Pr3

;------------------------------

;WM_CLOSE

Ty3PrY: INVOKE  DeleteObject,hImage3
        INVOKE  DestroyWindow,hWnd
        jmp     Ty3Pr3

;------------------------------

Type3Proc ENDP

;***********************************************************************************************

        OPTION  PROLOGUE:None
        OPTION  EPILOGUE:None

TableCreate PROC   hWndParent:HWND,lpTable:LPVOID

        push    esi
        push    edi
        mov     esi,[esp+16]      ;ESI = lpTable
        push    ebx
        push    ebp
        lodsd
        mov     edx,hInstance     ;EDX = hInstance
        xchg    eax,edi           ;EDI = lpHandleList
        lodsd
        mov     ebx,[esp+20]      ;EBX = hWndParent
        xchg    eax,ebp           ;EBP = table entry count

Creat0: lodsd           ;EAX = lParam
        push    edx     ;save hInstance
        push    eax                         ;CreateWindowEx:lParam
        lodsd           ;EAX = hMenu
        push    edx                         ;CreateWindowEx:hInstance
        push    eax                         ;CreateWindowEx:hMenu
        lodsd           ;EAX = nHeight
        push    ebx                         ;CreateWindowEx:hWndParent
        push    eax                         ;CreateWindowEx:nHeight
        lodsd           ;EAX = nWidth
        push    eax                         ;CreateWindowEx:nWidth
        lodsd           ;EAX = y (position)
        push    eax                         ;CreateWindowEx:y
        lodsd           ;EAX = x (position)
        push    eax                         ;CreateWindowEx:x
        lodsd           ;EAX = dwStyle
        push    eax                         ;CreateWindowEx:dwStyle
        lodsd           ;EAX = lpWindowName
        push    eax                         ;CreateWindowEx:lpWindowName
        lodsd           ;EAX = lpClassName
        push    eax                         ;CreateWindowEx:lpClassName
        lodsd           ;EAX = dwExStyle
        push    eax                         ;CreateWindowEx:dwExStyle
        CALL    CreateWindowEx
        or      edi,edi
        jz      Creat1

        stosd           ;store the handle

Creat1: mov     ecx,[esi-8]
        cmp     ecx,offset szBtnClass
        jz      Creat2

        cmp     ecx,offset szStcClass
        jnz     Creat3

Creat2: push    eax
        INVOKE  SendMessage,eax,WM_SETFONT,hFont,TRUE
        pop     eax

Creat3: dec     ebp
        pop     edx     ;EDX = hInstance
        jnz     Creat0

        pop     ebp
        pop     ebx
        pop     edi
        pop     esi
        ret     8

TableCreate ENDP

        OPTION  PROLOGUE:PrologueDef
        OPTION  EPILOGUE:EpilogueDef

;***********************************************************************************************

RegClass PROC

;EAX = hCursor
;EBX = hIcon
;ECX = hbrBackground
;EDX = lpfnWndProc
;ESI = lpszClassName
;EDI = 0
;EBP = hInstance

        push    ebx                                     ;WNDCLASSEX.hIconSm
        push    esi                                     ;WNDCLASSEX.lpszClassName
        push    edi                                     ;WNDCLASSEX.lpszMenuName
        push    ecx                                     ;WNDCLASSEX.hbrBackground
        push    eax                                     ;WNDCLASSEX.hCursor
        push    ebx                                     ;WNDCLASSEX.hIcon
        push    ebp                                     ;WNDCLASSEX.hInstance
        push    edi                                     ;WNDCLASSEX.cbWndExtra
        push    edi                                     ;WNDCLASSEX.cbClsExtra
        push    edx                                     ;WNDCLASSEX.lpfnWndProc
        push    edi                                     ;WNDCLASSEX.style
        push    sizeof WNDCLASSEX                       ;WNDCLASSEX.cbSize
        INVOKE  RegisterClassEx,esp
        add     esp,sizeof WNDCLASSEX
        ret

RegClass ENDP

;***********************************************************************************************

_main   PROC

;------------------------------

;initialize common controls

        xor     edi,edi                                 ;EDI = 0
        push    edi
        push    sizeof INITCOMMONCONTROLSEX
        INVOKE  InitCommonControlsEx,esp
        pop     ecx
        pop     edx

;------------------------------

;initialize hFont

        INVOKE  GetStockObject,ANSI_VAR_FONT
        mov     hFont,eax

;------------------------------

;initialize hInstance

        INVOKE  GetModuleHandle,edi
        mov     hInstance,eax
        xchg    eax,ebp                                 ;EBP = hInstance

;------------------------------

;register the window classes

        INVOKE  LoadIcon,ebp,IDI_ICON
        xchg    eax,ebx                                 ;EBX = hIcon
        INVOKE  LoadCursor,edi,IDC_ARROW                ;EAX = hCursor

;Type1

        mov     esi,offset szType1Class                 ;ESI = offset szClassName
        xor     ecx,ecx                                 ;ECX = hbrBackground
        mov     edx,Type1Proc                           ;EDX = lpfnWndProc
        push    eax
        call    RegClass
        pop     eax

;Type2

        mov     esi,offset szType2Class                 ;ESI = offset szClassName
        xor     ecx,ecx                                 ;ECX = hbrBackground
        mov     edx,Type2Proc                           ;EDX = lpfnWndProc
        push    eax
        call    RegClass
        pop     eax

;Type3

        mov     esi,offset szType3Class                 ;ESI = offset szClassName
        xor     ecx,ecx                                 ;ECX = hbrBackground
        mov     edx,Type3Proc                           ;EDX = lpfnWndProc
        push    eax
        call    RegClass
        pop     eax

;we do the main window last so that ESI is set up for CreateWindowEx

        mov     esi,offset szMainClass                  ;ESI = offset szClassName
        mov     ecx,COLOR_SCROLLBAR+1                   ;ECX = hbrBackground
        mov     edx,WndProc                             ;EDX = lpfnWndProc
        call    RegClass

;------------------------------

;create the main window

        INVOKE  CreateWindowEx,edi,esi,offset szAppName,
                WS_OVERLAPPEDWINDOW or WS_VISIBLE or WS_CLIPCHILDREN,
                CW_USEDEFAULT,SW_SHOWNORMAL,MainWinWidth,MainWinHeight,edi,edi,ebp,edi
        INVOKE  UpdateWindow,eax

;------------------------------

;message loop

        sub     esp,sizeof MSG
        mov     ebp,esp                                 ;EBP = addr MSG
        jmp short mLoop1

mLoop0: INVOKE  TranslateMessage,ebp
        INVOKE  DispatchMessage,ebp

mLoop1: INVOKE  GetMessage,ebp,edi,edi,edi
        inc     eax                                     ;exit only
        shr     eax,1                                   ;if 0 or -1
        jnz     mLoop0                                  ;otherwise, we loop

;------------------------------

;exit program

        INVOKE  ExitProcess,[ebp].MSG.wParam

;------------------------------

_main   ENDP

;###############################################################################################

        END     _main

